cell .baobab-cell-error {
    color: var(--error-color);
    font-weight: 700;
}

cell .baobab-cell-warning {
    color: var(--warning-color);
    font-weight: 700;
}

cell .baobab-cell-warning:selected,
cell .baobab-cell-error:selected {
    color: var(--accent-fg-color);
}

cell box.baobab-cell-box {
    border-spacing: 6px;
}

progresscell progress {
    background: var(--accent-bg-color);
    border-radius: 5px;
}

progresscell {
    border-radius: 5px;
    min-height: 18px;
}

progresscell trough {
    background: color-mix(in srgb, var(--border-color), transparent);
}

progresscell label {
    font-size: 0.9em;
    padding: 1px 3px;
}

progresscell label.in-progress {
    color: var(--accent-fg-color);
}

/* FolderDisplay */
window.background stack > grid > frame border {
    border-width: 0 1px 1px 0;
}

/* regular TreeView */
window.background stack > grid > scrolledwindow.frame {
    border-width: 0 1px 0 0;
}

window.background stack treeview:hover:not(:selected) {
    background: var(--window-bg-color);
}

.pathbar {
    border-radius: 6px;
    background-color: color-mix(in srgb, currentColor 10%, transparent);
}

.pathbar button {
    margin: 3px;
    border-radius: 4px;
    padding-top: 0px;
    padding-bottom: 0px;
}

.pathbar button:not(:hover),
.pathbar button.current-dir {
  background: none;
  box-shadow: none;
}

.pathbar button:not(.current-dir):not(:backdrop):hover label,
.pathbar button:not(.current-dir):not(:backdrop):hover image {
  opacity: 1;
}

.pathbar > scrolledwindow undershoot.left {
  background: linear-gradient(to right, var(--headerbar-shade-color) 6px, transparent 24px);
}

.pathbar > scrolledwindow undershoot.right {
  background: linear-gradient(to left, var(--headerbar-shade-color) 6px, transparent 24px);
}

image.symbolic-circular {
  background-color: color-mix(in srgb, currentColor 8%, transparent);
  min-width: 32px;
  min-height: 32px;
  border-radius: 50%;
}

window infobar.warning,
window infobar.error {
    border-bottom-width: 1px;
    border-bottom-style: solid;
    border-bottom-color: var(--border-color);
}

ringschart.subfolder-tip {
    color: white;
    background-color: rgb(0 0 0 / 0.8);
    background-clip: padding-box;
    border-radius: 9px;
    padding: 6px 10px;
}

:drop(active) .contents {
  outline-style: solid;
  outline-width: 3px;
  outline-color: color-mix(in srgb, var(--accent-bg-color), transparent);
  outline-offset: -3px;
}
